import numpy as np
from scipy.integrate import trapz

from .orientations import zcw_spherical as crystallites


class Powder(object):
    def __init__(self, n=100):
        self.alpha, self.beta, self.wt = crystallites(n)


class Pake(object):
    parameter = {'delta': 0., 'eta': 0., 'w_iso': 0.,
                 'x': None, 'y': None, 'half': False}

    def __init__(self, samples=10000):
        self.powder = Powder(n=samples)
        self.dw = None
        self.parameter = {}
        self.GB = 1000
        self.pw = 1e-6
        self.legendre = 0
        self.axial = 0
        self.calc_prefactor()

    def calc_prefactor(self):
        """
        Calculates angular dependencies
        Returns
        -------

        """
        self.legendre = 3 * np.square(np.cos(self.powder.beta)) - 1
        self.axial = np.square(np.sin(self.powder.beta)) * np.cos(2 * self.powder.alpha)

    def set_parameter(self, p, dw=None):
        self.parameter['delta'] = p[0] * 1e3
        self.parameter['eta'] = p[1]
        self.GB = p[2] * 1e3
        self.pw = p[3] * 1e-6
        if len(p) == 5:
            self.parameter['w_iso'] = p[4] * 1e3
        if dw is not None:
            self.dw = dw

    @staticmethod
    def _check_old_data(x, d, e, w, h):
        if Pake.parameter['x'] is None:
            return False
        if Pake.parameter['w_iso'] == w and Pake.parameter['delta'] == d and \
                        Pake.parameter['eta'] == e and Pake.parameter['half'] == h and \
                        Pake.parameter['x'][0] == x[0] and Pake.parameter['x'][1] == x[1]:
            return True
        return False

    def calc_timesignal(self, tt, p, half=False):
        """
        Calculate time signal of pake spectrum
        Parameters
        ----------
        tt: time array
        p: list of parameters [delta, eta, GB, omega_iso]
        half: if False, Pake spectrum will be calculated, if True, chemical shift spectrum will be calculated

        Returns
        -------
        timesignal
        """
        _delta, _eta, _wiso = p
        omega = 0.5 * _delta * (self.legendre - _eta * self.axial)
        if half:
            omega += _wiso
        weight = self.powder.wt
        if half:
            fid_real = np.zeros_like(tt, dtype=float)
            fid_imag = np.zeros_like(tt, dtype=float)
            for k, v in zip(omega, weight):
                pha = 2 * np.pi * k * tt
                fid_real += (np.cos(pha)) * v
                fid_imag += (np.sin(pha)) * v
            fid = 1j * fid_imag
            fid += fid_real
        else:
            fid = np.zeros_like(tt, dtype=float)
            for k, v in zip(omega, weight):
                fid += np.cos(2 * np.pi * k * tt) * v
        return fid

    @staticmethod
    def calc_pulse(pw, f):
        _phi = 2 * np.pi * f * pw
        pulse = ((0.5 * np.pi * np.sin(np.sqrt(0.25 * np.pi ** 2 + 0.5 * _phi ** 2))) /
                 np.sqrt(0.25 * np.pi ** 2 + 0.25 * _phi ** 2)) ** 1

        return pulse

    def pake_spectrum(self, x, p=None, half=False):
        if p is None:
            _gb = self.GB
            _pw = self.pw
            try:
                _delta = self.parameter['delta']
                _eta = self.parameter['eta']
                _wiso = self.parameter['w_iso']
            except KeyError:
                _delta = Pake.parameter['delta']
                _eta = Pake.parameter['eta']
                _wiso = Pake.parameter['w_iso']
        else:
            _delta = p[0] * 1e3
            _eta = p[1]
            _gb = p[2] * 1e3
            _pw = p[3] * 1e-6
            _wiso = p[4] * 1e3

        # x is time
        if abs(x[1] - x[0]) < 1:
            if self.dw is None:
                self.dw = x[1] - x[0]
            t = x
            nop = len(x)
            isfreq = False
        # x is frequency
        else:
            df = (x[1] - x[0])
            if self.dw is None:
                self.dw = 1 / (df * len(x))
            if len(x) != 1 / (self.dw * df):
                nop = int(1 / (self.dw * df))
            else:
                nop = len(x)

            # avoid calculation for to much points
            t = np.arange(0, min(8192, nop)) * self.dw
            isfreq = True

        new_data = self._check_old_data((nop, self.dw), _delta, _eta, _wiso, half)

        if new_data:
            fid = Pake.parameter['y'] + 0
            Pake.parameter['delta'] = _delta
            Pake.parameter['eta'] = _eta
            Pake.parameter['w_iso'] = _wiso
            Pake.parameter['x'] = (nop, self.dw)
            Pake.parameter['half'] = half
            self.GB = _gb
            self.pw = _pw
            fid = self.calc_timesignal(t, [_delta, _eta, _wiso], half=half)
            Pake.parameter['y'] = fid + 0


        fid *= np.exp(-(_gb * t) ** 2)

        # add zeros to get desired signal length
        if len(fid) != nop:
            fid = np.append(fid, np.zeros((nop - len(fid))), axis=0)

        spec = np.fft.fftshift(np.fft.fft(fid))
        spec -= np.mean(spec[:10])
        _f = np.fft.fftshift(np.fft.fftfreq(nop, d=self.dw))

        # take finite pulse length into account
        if _pw != 0:
            pulse = self.calc_pulse(_pw, _f)
            spec *= pulse

        # shift pake spectrum by isotropic shift
        if not half:
            spec = np.roll(spec, int(self.dw * _wiso * nop))

        # cut to original frequency range
        if isfreq:
            # mask = np.ma.masked_inside(_f, min(x), max(x)).mask
            # spec = spec[mask]
            spec = np.interp(x, _f, spec.real)

        spec.real -= np.mean(spec.real[:10])

        return spec.real / max(spec.real)

    def pake(self, x, p=None):
        return self.pake_spectrum(x, p=p)

    def halfpake(self, x, p=None):
        return self.pake_spectrum(x, p=p, half=True)


class SecondOrder(object):
    parameter = {'c_q': 1e6, 'eta': 0.,
                 'x': None, 'y': None}

    def __init__(self, samples=200000):
        self.powder = Powder(n=samples)
        self.dw = None
        self.parameter = {}
        self.GB = 1000
        self.pw = 1e-6
        self.legendre = 0
        self.axial = 0
        self.coupling = 1
        self.larmor = 2 * np.pi * 100e6

    def set_parameter(self, p, dw=None, larmor=None, spin=2.5):
        self.parameter['c_q'] = 2 * np.pi * p[0] * 1e6
        self.parameter['eta'] = p[1]
        self.GB = p[2] * 1e3
        if dw is not None:
            self.dw = dw
        if larmor is not None:
            self.larmor = 2 * np.pi * larmor
        self._calc_coupling(2 * np.pi * p[0] * 1e6, spin=spin)

    def _calc_coupling(self, c_q, spin=2.5):
        omega_q = c_q / (2 * spin * (2 * spin - 1))
        self.coupling = 1.5 * omega_q ** 2 / self.larmor * (spin * (spin + 1) - 0.75)

    def calc_omega(self, eta, delta):
        self._calc_coupling(delta)
        cos2phi = np.cos(2 * self.powder.alpha)
        cos_theta_square = 0.5 + 0.5 * np.cos(2 * self.powder.beta)
        prefactor_a = -3.375 + 2.25 * eta * cos2phi - 0.375 * (eta * cos2phi) ** 2
        prefactor_b = 3.75 - 0.5 * eta ** 2 - 2 * eta * cos2phi + 0.75 * (eta * cos2phi) ** 2
        prefactor_c = -0.375 + (eta ** 2) / 3. - 0.25 * eta * cos2phi - 0.375 * (eta * cos2phi) ** 2
        ret_val = np.zeros_like(cos2phi)
        ret_val += prefactor_a * cos_theta_square ** 2
        ret_val += prefactor_b * cos_theta_square
        ret_val += prefactor_c
        return self.coupling * ret_val

    def calc_timesignal(self, tt, p):
        delta, eta = p
        omega = self.calc_omega(eta, delta)
        weight = self.powder.wt
        fid_real = np.zeros_like(tt, dtype=float)
        fid_imag = np.zeros_like(tt, dtype=float)

        for o, w in zip(omega, weight):
            pha = o * tt
            co_pha = np.cos(pha)
            si_pha = np.sin(pha)
            fid_real += co_pha * w
            fid_imag += si_pha * w

        fid = 1j * fid_imag
        fid += fid_real
        return fid

    def spectrum(self, x, p=None):
        if p is None:
            p = [self.parameter['c_q'], self.parameter['eta'], self.GB]
        if abs(x[1] - x[0]) < 1:
            if self.dw is None:
                self.dw = x[1] - x[0]
            t = x
            nop = len(x)
            isfreq = False
        else:
            df = (x[1] - x[0])
            if self.dw is None:
                self.dw = 1 / (df * len(x))
            if len(x) != 1 / (self.dw * df):
                nop = int(1 / (self.dw * df))
            else:
                nop = len(x)
            t = np.arange(0, min(8192, nop)) * self.dw
            isfreq = True

        fid = self.calc_timesignal(t, p[:-1])
        fid *= np.exp(-(p[-1] * t) ** 2)

        if len(fid) != nop:
            fid = np.append(fid, np.zeros((nop - len(fid))), axis=0)

        spec = np.fft.fftshift(np.fft.fft(fid))
        _f = np.fft.fftshift(np.fft.fftfreq(nop, d=self.dw))

        if isfreq:
            mask = np.ma.masked_inside(_f, min(x), max(x)).mask
            spec = spec[mask]

        return spec.real - np.mean(spec.real[:10])


def csa(x: np.ndarray, delta: float, eta: float, wiso: float, gb: float, sw: float):
    dw = 1 / sw
    a, b, _ = crystallites(200000)
    bins = 0.5 * (x[1:] + x[:-1])
    bins = np.hstack((np.array([0.5 * (-x[1] + 3 * x[0])]), bins, np.array([0.5 * (3 * x[-1] - x[-2])])))

    omega = wiso + delta * 0.5 * (3 * np.cos(b) ** 2 - 1 - eta * np.square(np.sin(b)) * np.cos(2 * a))

    s_left = np.histogram(omega, bins=bins)[0]
    s = s_left

    if gb != 0:
        apd = np.exp(-4 * np.log(2) * (np.fft.fftshift(np.fft.fftfreq(len(x), d=dw)) / sw) ** 2) * \
              2 * np.sqrt(np.log(2) / np.pi) / gb
        ret_val = np.convolve(s, apd, mode='same')
    else:
        ret_val = s

    return ret_val / trapz(ret_val, x)


def pake(p, x):
    dw = 1/p[-1]
    a, b, _ = crystallites(200000)
    bins = 0.5 * (x[1:] + x[:-1])
    bins = np.hstack((np.array([0.5 * (-x[1] + 3 * x[0])]), bins, np.array([0.5 * (3 * x[-1] - x[-2])])))

    omega = p[0] * 0.5 * (3*np.cos(b)**2 - 1 - p[1] * np.square(np.sin(b)) * np.cos(2 * a))

    s_left = np.histogram(omega, bins=bins)[0]
    s_right = np.histogram(-omega, bins=bins)[0]
    s = s_left + s_right

    if p[2] != 0:
        apd = np.exp(-4 * np.log(2) * (np.fft.fftshift(np.fft.fftfreq(len(x), d=dw)) / p[2]) ** 2) * \
              2 * np.sqrt(np.log(2) / np.pi) / p[2]
        ret_val = np.convolve(s, apd, mode='same')
    else:
        ret_val = s

    return ret_val/trapz(ret_val, x)


def sec_order(p, x, spin=2.5):
    dw = 1/p[-1]
    a, b, _ = crystallites(200000)
    bins = 0.5 * (x[1:] + x[:-1])
    bins = np.hstack((np.array([0.5 * (-x[1] + 3 * x[0])]), bins, np.array([0.5 * (3 * x[-1] - x[-2])])))

    omega_q = 2 * np.pi * p[0] / (2*spin * (2*spin - 1))
    coupling = 1.5 * (omega_q**2 / (2 * np.pi *p[3])) * (spin*(spin + 1) - 0.75)

    cos2phi = np.cos(2 * a)
    cos_theta_square = 0.5+0.5*np.cos(2*b)
    prefactor_a = -3.375 + 2.25*p[1] * cos2phi - 0.375 * (p[1] * cos2phi)**2
    prefactor_b = 3.75 - 0.5 * p[1]**2 - 2*p[1]*cos2phi + 0.75 * (p[1]*cos2phi)**2
    prefactor_c = -0.375 + (p[1]**2)/3. - 0.25*p[1]*cos2phi - 0.375 * (p[1]*cos2phi)**2
    orient = np.zeros_like(cos2phi)
    orient += prefactor_a * cos_theta_square ** 2
    orient += prefactor_b * cos_theta_square
    orient += prefactor_c

    omega = coupling * orient

    s = np.histogram(omega/(2*np.pi), bins=bins)[0]

    if p[2] != 0:
        apd = np.exp(-4 * np.log(2) * (np.fft.fftshift(np.fft.fftfreq(len(x), d=dw)) / p[2]) ** 2) * \
              2 * np.sqrt(np.log(2) / np.pi) / p[2]
        ret_val = np.convolve(s, apd, mode='same')
    else:
        ret_val = s

    return ret_val / trapz(ret_val, x)


if __name__ == '__main__':
    # pass
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    dw0 = 4e-6
    xxx = np.arange(0, 8192) * dw0

    freq = np.fft.fftshift(np.fft.fftfreq(len(xxx), d=dw0))[500:-1000]

    p0 = [6.66e6, 0.935, 2e3, 50e6, 1/dw0]
    spec = sec_order(p0, freq)
    ax.plot(freq, spec)

    p1 = [20e3, 0.2, 2e3, 1/dw0]
    spec1 = pake(p1, freq)
    ax.plot(freq, spec1)

    p2 = [20e3, 0.2, 20e3, 2e3, 1/dw0]
    spec2 = csa(p2, freq)
    ax.plot(freq, spec2)

    plt.show()
