
import numpy as np
from scipy.special import gammaln
from ..math.logfourier import logft

stepsize = 0.2  # 0.05
span1 = 20
span2 = 30


class GeneralGamma(object):
    lower_b = 20
    upper_b = 20
    dx = 0.2

    @staticmethod
    def gga_dist(tau, tau0, alpha, beta):
        b_to_a = beta / alpha
        norm = np.exp(gammaln(b_to_a) - b_to_a * np.log(b_to_a)) / alpha
        t_to_t0 = tau / tau0
        ret_val = np.exp(-b_to_a * t_to_t0 ** alpha) * t_to_t0 ** beta

        return ret_val / norm

    @staticmethod
    def gga_ew_dist(tau, tau0, alpha, beta, sigma, gamma):
        if gamma == beta:
            return GeneralGamma.gga_dist(tau, tau0, alpha, beta)
        b_to_a = beta / alpha
        g_to_a = gamma / alpha
        t_to_t0 = tau / tau0
        norm = (np.exp(gammaln(b_to_a)) + sigma**(gamma-beta) * np.exp(gammaln(g_to_a) + (b_to_a-g_to_a)*np.log(b_to_a))) / np.exp(b_to_a*np.log(b_to_a)) / alpha

        ret_val = np.exp(-b_to_a * t_to_t0**alpha) * t_to_t0**beta * (1 + (t_to_t0*sigma)**(gamma-beta))

        return ret_val / norm

    @staticmethod
    def ggb_dist(tau, tau0, a, b):
        norm = a * (1+b) * np.sin(np.pi*b/(1+b)) * b**(b/(1+b)) / np.pi
        ret_val = b * (tau/tau0)**a + (tau/tau0)**(-a*b)

        return norm / ret_val

    @staticmethod
    def _gg_time(t, tau0, *args, **kwargs):
        logtau = np.log(tau0)
        steps = np.arange(logtau - GeneralGamma.lower_b, logtau + GeneralGamma.upper_b, GeneralGamma.dx)
        taus = np.exp(steps)
        ret_val = np.zeros_like(t)
        dist = {'gga': GeneralGamma.gga_dist,
                'gga_ew': GeneralGamma.gga_ew_dist,
                'ggb': GeneralGamma.ggb_dist}[kwargs.pop('mode', 'gga')]
        gtau = dist(taus, tau0, *args)
        for i, xval in enumerate(t):
            y = np.exp(- xval / taus)
            y *= gtau
            y = y[1:] + y[:-1]
            ret_val[i] = np.sum(y) * GeneralGamma.dx / 2
        return ret_val

    @staticmethod
    def ggaew_b_freq(f, tau0, alpha, beta, sigma, gamma, tau1, a, b, r, dec=5):
        t = np.logspace(np.log10(1 / max(f)) - dec, np.log10(1 / min(f)) + dec, num=int(len(f) * (0.5 * dec)))
        phi_a = GeneralGamma.ggaew(t, tau0, alpha, beta, sigma, gamma)
        phi_b = GeneralGamma.ggb(t, tau1, a, b)
        temp = phi_a * (1 - r + r * phi_b)

        ret_val = logft(t, temp, new_x=2 * np.pi * f)[1].y

        return ret_val * f * 2 * np.pi

    @staticmethod
    def gga_b_freq(f, tau0, alpha, beta, tau1, a, b, r, dec=5):
        t = np.logspace(np.log10(1 / max(f)) - dec, np.log10(1 / min(f)) + dec, num=int(len(f) * (0.5 * dec)))
        phi_a = GeneralGamma.gga(t, tau0, alpha, beta)
        phi_b = GeneralGamma.ggb(t, tau1, a, b)
        temp = phi_a * (1 - r + r * phi_b)

        ret_val = logft(t, temp, new_x=2 * np.pi * f)[1].y

        return ret_val * f * 2 * np.pi

    @staticmethod
    def gga(t, tau0, alpha, beta):
        return GeneralGamma._gg_time(t, tau0, alpha, beta, mode='gga')

    @staticmethod
    def ggaew(t, tau0, alpha, beta, sigma, gamma):
        return GeneralGamma._gg_time(t, tau0, alpha, beta, sigma, gamma, mode='gga_ew')

    @staticmethod
    def ggb(t, tau0, a, b):
        return GeneralGamma._gg_time(t, tau0, a, b, mode='ggb')

    @staticmethod
    def ggaew_b(t, tau0, alpha, beta, sigma, gamma, tau1, a, b, r):
        phi_a = GeneralGamma.ggaew(t, tau0, alpha, beta, sigma, gamma)
        phi_b = GeneralGamma.ggb(t, tau1, a, b)
        ret_val = phi_a * (1 - r + r * phi_b)
        return ret_val

    @staticmethod
    def gga_b(t, tau0, alpha, beta, tau1, a, b, r):
        phi_a = GeneralGamma.gga(t, tau0, alpha, beta)
        phi_b = GeneralGamma.ggb(t, tau1, a, b)
        ret_val = phi_a * (1 - r + r * phi_b)
        return ret_val


class FitGG(object):
    name = 'Alpha + Beta + EW'
    type = 'Gamma-Functions'
    equation = r'A*\Phi_{\alpha,EW} * [1 - C + C*\Phi_{\beta}] + B'
    params = ['A', 'B',
              r'\tau_{\alpha}', r'\alpha', r'\beta', '\sigma', '\gamma',
              r'\tau_{\beta}', 'a', 'b', 'C']

    @staticmethod
    def func(p, x):
        amp, base, tau0, alpha, beta, sigma, gamma, tau1, a, b, c = p
        phi_a = GeneralGamma.ggaew(x, tau0, alpha, beta, sigma, gamma)
        phi_b = GeneralGamma.ggb(x, tau1, a, b)
        ret_val = amp * phi_a * (1 - c + c * phi_b) + base
        return ret_val


# Definiere Funktionen #
def glntau1_td(ttime, tau, alpha, beta):
    lnx = np.log(tau) + span1
    intmin = np.log(tau) - span2
    steps = np.arange(lnx, intmin, -stepsize)
    explnx = np.exp(steps)
    betatoalpha = beta / alpha
    logtau = np.log(tau)
    if betatoalpha > 80:
        norm = np.exp(.5 * np.log(2 * np.pi / betatoalpha) - betatoalpha) / alpha
    else:
        norm = np.exp(gammaln(betatoalpha)) / alpha / np.power(betatoalpha, betatoalpha)
    y = np.exp(-np.exp((steps - logtau) * alpha) * betatoalpha) * np.power(explnx / tau, beta) * np.exp(-ttime / explnx)
    y = y[1:] + y[:-1]
    intval = np.sum(y) / 2
    intval = intval * stepsize / norm
    return intval


def gga(t, tau0, alpha, beta):
    logtau = np.log(tau0)
    stepsize = 0.2
    steps = np.arange(logtau - 20, logtau + 20, stepsize)
    taus = np.exp(steps)/tau0
    betatoalpha = beta / alpha
    norm = np.exp(gammaln(betatoalpha) - betatoalpha*np.log(betatoalpha)) / alpha
    ret_val = np.zeros_like(t)
    gtau = np.exp(-taus ** alpha * betatoalpha) * taus ** beta
    for i, xval in enumerate(t):
        # exponent = -taus**alpha * betatoalpha - xval/taus/tau0
        y = np.exp(- xval/taus/tau0)
        y *= gtau
        y = y[1:] + y[:-1]
        ret_val[i] = np.sum(y) * stepsize / norm / 2
    return ret_val


def ggaew(t, tau0, alpha, beta, sigma, gamma):
    logtau = np.log(tau0)
    stepsize = 0.2
    steps = np.arange(logtau - 20, logtau + 20, stepsize)
    taus = np.exp(steps) / tau0
    q1 = beta / alpha
    q2 = gamma / alpha
    norm = (np.exp(gammaln(q1)) + sigma**(gamma-beta) * np.exp((q1-q2)*np.log(q1)) * np.exp(gammaln(q2))) / alpha / np.exp(q1*np.log(q1))
    ret_val = np.empty_like(t)
    gtau = np.exp(-taus**alpha * q1) * taus**beta * (1 + (taus*sigma)**(gamma-beta))
    for i, xval in enumerate(t):
        y = np.exp(-xval/taus/tau0)
        y *= gtau
        y = y[1:] + y[:-1]
        ret_val[i] = np.sum(y) * stepsize / norm / 2
    return ret_val


def ggaew_b(t, tau0, alpha, beta, sigma, gamma, tau1, a, b, r):
    phi_a = ggaew(t, tau0, alpha, beta,sigma, gamma)
    phi_b = ggb(t, tau1, a, b)
    ret_val = phi_a * (1-r + r*phi_b)
    return ret_val


def ggb(t, tau0, a, b):
    logtau = np.log(tau0)
    stepsize = 0.2
    steps = np.arange(logtau - 20, logtau + 20, stepsize)
    taus = np.exp(steps) / tau0
    norm = np.pi / (a*(1+b)) / b**(b/(1+b)) / np.sin(np.pi*b / (1+b))
    ret_val = np.empty_like(t)
    gtau = 1. / (b * taus**a + taus**(-b*a))
    for i, xval in enumerate(t):
        y = np.exp(-xval / taus / tau0)
        y *= gtau
        y = y[1:] + y[:-1]
        ret_val[i] = np.sum(y) * stepsize / norm / 2
    return ret_val


def glntau2_td(ttime, tau, alpha, beta, sigma, gam):
    intval = 0
    lnx = np.log(tau) + span1
    intmin = np.log(tau) - span2
    steps = np.arange(lnx, intmin, -stepsize)
    explnx = np.exp(steps)
    logtau = np.log(tau)
    q1 = beta / alpha
    q2 = gam / alpha

    if q2 > 80:
        norm = (np.exp(.5 * np.log(2 * np.pi / q1) - q1) + np.power(sigma, alpha * (q2 - q1)) * np.power(q2 / q1,
                                                                                                      -q2) * np.exp(
            .5 * np.log(2 * np.PI / q2) - q2)) / alpha
    elif q1 > 80:
        norm = (np.exp(.5 * np.log(2 * np.pi / q1) - q1) + np.power(sigma, alpha * (q2 - q1)) * np.power(q2 / q1,
                                                                                                    -q2) * np.exp(
            gammaln(q2)) * np.power(q1, -q1)) / alpha
    else:
        norm = (np.exp(gammaln(q1)) + np.power(sigma, gam - beta) * np.power(q1, q1 - q2) * np.exp(gammaln(q2))) / (
                alpha * np.power(q1, q1))
    y = np.exp(-np.exp((steps - logtau) * alpha) * q1) * np.power(explnx / tau, beta) * (
            1 + np.power(explnx * sigma / tau, gam - beta)) * np.exp(-ttime / explnx)
    y = y[1:] + y[:-1]
    intval = np.sum(y) / 2
    intval = intval * stepsize / norm
    return intval


def glnbeta_td(ttime, tau, a, b):
    intval = 0
    lnx = np.log(tau) + span1
    intmin = np.log(tau) - span2
    steps = np.arange(lnx, intmin, -stepsize)
    explnx = np.exp(steps)
    norm = np.pi / (a * (1 + b)) / np.power(b, b / (1. + b)) / np.sin(np.pi * b / (1. + b))
    y = 1. / (b * np.power(explnx / tau, a) + np.power(explnx / tau, -(b * a))) * np.exp(-ttime / explnx)
    y = y[1:] + y[:-1]
    intval = np.sum(y) / 2

    intval = intval * stepsize / norm
    return intval


def g2(p, x):
    ret_val = np.empty(np.shape(x))
    for i, xval in enumerate(x):
        # fkt1 = glntau2_td(xval, p[1], p[2], p[3], p[4], p[5])
        # fkt1 = glntau1_td(xval, p[1], p[2], p[3])
        fkt1 = glntau2_td(xval, p[1], p[2], p[3], p[4], p[5])
        fkt2 = glnbeta_td(xval, p[6], p[7], p[8])
        # ret_val[i] = (fkt1 * ((1 - p[4]) + p[4] * fkt2) * p[0]) + np.abs(p[9]) * np.exp(-(xval/p[10])**p[11]) + p[8]
        ret_val[i] = fkt1 * (1-p[0] + p[0]*fkt2)
    return ret_val


def g1(p, x):
    ret_val = np.empty(np.shape(x))
    for i, xval in enumerate(x):
        #        fkt1 = glntau1_td(xval,p[1],p[2],p[3])
        fkt1 = glntau2_td(xval, p[1], p[2], p[3], p[11], p[12])
        fkt2 = glnbeta_td(xval, p[5], p[6], p[7])
        ret_val[i] = p[0] * (fkt1 * ((1 - p[4]) + p[4] * fkt2)) + np.abs(p[8]) * np.exp(-(xval / p[9]) ** p[10])
    return ret_val


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import time

    x = np.logspace(-5, 5, num=260)
    tau0 = 100
    alpha = 2
    beta = 0.7
    gamma = 0.1
    sigma = 1000
    tau1 = 1
    a = 0.8
    b = 0.6
    r = 0
    start = time.time()
    y = GeneralGamma.ggaew_b(x, tau0, alpha, beta, sigma, gamma, tau1, a, b, r)
    print('gg_time', time.time()-start)
    start = time.time()
    yy = g2([r, tau0, alpha, beta, sigma, gamma, tau1, a, b], x)
    print('g2', time.time()-start)

    plt.semilogx(x, y)
    plt.semilogx(x, yy, 'x')
    plt.show()



